package de.jottyfan.timetrack.modules.done;

import java.io.IOException;
import java.io.Serializable;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TimeZone;

import javax.enterprise.context.SessionScoped;
import javax.faces.application.FacesMessage;
import javax.inject.Named;

import org.jooq.exception.DataAccessException;

import de.jooqfaces.JooqFacesContext;
import de.jottyfan.timetrack.db.done.tables.records.TBillingRecord;
import de.jottyfan.timetrack.db.done.tables.records.TJobRecord;
import de.jottyfan.timetrack.db.done.tables.records.TModuleRecord;
import de.jottyfan.timetrack.db.done.tables.records.TProjectRecord;
import de.jottyfan.timetrack.modules.Model;

/**
 * 
 * @author henkej
 *
 */
@Named
@SessionScoped
public class DoneModel implements Model, Serializable {
	private static final long serialVersionUID = 1L;

	private DoneBean bean;
	private List<DoneBean> beans;
	private List<TProjectRecord> projects;
	private List<TModuleRecord> modules;
	private List<TJobRecord> activities;
	private List<TBillingRecord> billings;
	private List<TimeBean> times;
	private List<DailySummaryBean> allJobs;
	private WholeDaySummaryBean daySummary;
	private java.util.Date day;
	private String calendarEvents;
	private List<UserBean> users;
	private DownloadBean downloadBean;

	public boolean init(JooqFacesContext facesContext) {
		try {
			day = day == null ? new java.util.Date() : day;
			beans = getAllOfDay(facesContext, day);
			DoneGateway gw = new DoneGateway(facesContext);
			modules = gw.getAllModules();
			activities = gw.getAllActivities();
			projects = gw.getAllProjects();
			billings = gw.getAllBillings();
			daySummary = gw.getDaySummary(day);
			allJobs = gw.getAllJobs(day);
			calendarEvents = gw.getAllCalendarEvents();
			return true;
		} catch (DataAccessException | ClassNotFoundException | SQLException e) {
			facesContext.addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "error", e.getMessage()));
			return false;
		}
	}

	public boolean loadDefaults(JooqFacesContext facesContext) {
		try {
			defineTimes();
			return true;
		} catch (DataAccessException e) {
			facesContext.addMessage(null,
					new FacesMessage(FacesMessage.SEVERITY_ERROR, "error loading defaults", e.getMessage()));
			return false;
		}
	}

	public boolean prepareRead(JooqFacesContext facesContext) {
		try {
			users = new DoneGateway(facesContext).getAllUsers();
			downloadBean = new DownloadBean();
			downloadBean.setFromDate(LocalDate.of(LocalDate.now().getYear(), 1, 1));
			downloadBean.setUntilDate(LocalDate.now());
			return true;
		} catch (DataAccessException | ClassNotFoundException | SQLException e) {
			facesContext.addMessage(null,
					new FacesMessage(FacesMessage.SEVERITY_ERROR, "error loading defaults", e.getMessage()));
			return false;
		}
	}

	private void defineTimes() {
		times = new ArrayList<>();
		Integer currentHour = getCurrentTime().get(GregorianCalendar.HOUR_OF_DAY);
		Integer currentMinute = getCurrentTime().get(GregorianCalendar.MINUTE);
		if (currentMinute < 15) {
			currentMinute = 0;
		} else if (currentMinute < 30) {
			currentMinute = 15;
		} else if (currentMinute < 45) {
			currentMinute = 30;
		} else {
			currentMinute = 45;
		}
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 120), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 105), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 90), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 75), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 60), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 45), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 30), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute - 15), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute), "default"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute + 15), "default"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute + 30), "link"));
		times.add(new TimeBean(defineNewTime(currentHour, currentMinute + 45), "link"));
	}
	
	/**
	 * summarizes the billings in horizon2020 notation
	 * 
	 * @return the billings summary
	 */
	public String getSummaryBilling() {
		Map<String, Double> map = new HashMap<>();
		for (DailySummaryBean bean : getAllJobs()) {
			String shortcut = bean.getBillingShortcut();
			Double time = bean.getDurationHours();
			Double sumTime = map.get(shortcut);
			sumTime = sumTime == null ? time : (sumTime + time);
			map.put(shortcut, sumTime);
		}
		StringBuilder buf = new StringBuilder();
		boolean first = true;
		for (Entry<String, Double> entry : map.entrySet()) {
			buf.append(first ? "" : ", ");
			String key = entry.getKey();
			if (key == null || key.equals("null")) {
				buf.append("$ ");
			} else {
				buf.append(entry.getKey()).append(" ");
			}
			buf.append("(").append(entry.getValue()).append(")");
			first = false;
		}
		return buf.toString();
	}
	
	/**
	 * summarizes the time of a days work in commaseparated values
	 * 
	 * @return the summarized times
	 */
	public String getSummaryTime() {
		Double sum = 0d;
		for (DailySummaryBean bean : getAllJobs()) {
			sum += bean.getDurationHours();
		}
		return sum.toString();
	}

	private Calendar getCurrentTime() {
		return new GregorianCalendar(TimeZone.getTimeZone("Europe/Berlin"));
	}

	private Calendar defineNewTime(Integer hours, Integer minutes) {
		Calendar cal = new GregorianCalendar();
		if (minutes < 0) {
			hours--;
			minutes += 60;
		} else if (minutes > 59) {
			hours++;
			minutes -= 60;
		}
		cal.set(0, 0, 0, hours, minutes, 0);
		return cal;
	}

	/**
	 * get all entries of t_done
	 * 
	 * @param facesContext
	 * @param day
	 *          the day to look up for
	 * @param login
	 *          the user to look up for
	 * @return all entries
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	private List<DoneBean> getAllOfDay(JooqFacesContext facesContext, Date day)
			throws DataAccessException, ClassNotFoundException, SQLException {
		LocalDateTime ldt = day.toInstant().atZone(ZoneId.systemDefault()).toLocalDateTime();
		return new DoneGateway(facesContext).getAll(ldt);
	}

	public boolean insert(JooqFacesContext facesContext) {
		try {
			bean.setDay(day);
			new DoneGateway(facesContext).insert(bean);
			return true;
		} catch (DataAccessException | ClassNotFoundException | SQLException e) {
			facesContext.addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "error", e.getMessage()));
			return false;
		}
	}

	public boolean update(JooqFacesContext facesContext) {
		try {
		  // day contains the correct date; with that, refresh the current date after possibly changing the day
		  bean.setDay(day);
			new DoneGateway(facesContext).update(bean);
			return true;
		} catch (DataAccessException | ClassNotFoundException | SQLException e) {
			facesContext.addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "error", e.getMessage()));
			return false;
		}
	}

	public boolean delete(JooqFacesContext facesContext) {
		try {
			new DoneGateway(facesContext).delete(bean);
			return true;
		} catch (DataAccessException | ClassNotFoundException | SQLException e) {
			facesContext.addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "error", e.getMessage()));
			return false;
		}
	}

	public boolean download(JooqFacesContext facesContext) {
		try {
			String csvContent = new DoneGateway(facesContext).download(downloadBean);
			facesContext.getExternalContext().responseReset();
			facesContext.getExternalContext().setResponseContentType("text/csv");
			facesContext.getExternalContext().setResponseHeader("Content-Disposition", "attachment; filename=\"worktime.csv\"");
			facesContext.getExternalContext().getResponseOutputWriter().append(csvContent);
			facesContext.responseComplete();
			return true;
		} catch (DataAccessException | ClassNotFoundException | SQLException | IOException e) {
			facesContext.addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "error", e.getMessage()));
			return false;
		}
	}

	public String getAttach() {
		StringBuilder buf = new StringBuilder();
		String thatday = new SimpleDateFormat("dd.MM.yyyy").format(day);
		for (DailySummaryBean sdb : allJobs) {
			buf.append(thatday).append("\t");
			buf.append(sdb.getDuration()).append("\t");
			buf.append(sdb.getProjectName());
			if (sdb.getBillingShortcut() != null && !sdb.getBillingShortcut().isBlank()) {
			  buf.append(" (").append(sdb.getBillingShortcut()).append(")");
			}
			buf.append("\t");
			buf.append(sdb.getModuleName()).append("\t");
			buf.append(sdb.getJobName()).append("\t");
			buf.append("\n");
		}
		return buf.toString();
	}

	public String getDayIso8601() {
		return day == null ? "" : new SimpleDateFormat("yyyy-MM-dd").format(day);
	}

	public void setBean(DoneBean bean) {
		this.bean = bean;
	}

	public DoneBean getBean() {
		return bean;
	}

	public List<TProjectRecord> getProjects() {
		return projects;
	}

	public List<TModuleRecord> getModules() {
		return modules;
	}

	public List<TJobRecord> getActivities() {
		return activities;
	}
	
	public List<TBillingRecord> getBillings() {
		return billings;
	}

	public List<DailySummaryBean> getAllJobs() {
		return allJobs;
	}

	public WholeDaySummaryBean getDaySummary() {
		return daySummary;
	}

	public List<DoneBean> getBeans() {
		return beans;
	}

	public List<TimeBean> getTimes() {
		return times;
	}

	public String getGermanFormattedDay() {
	  return day == null ? "null" : new SimpleDateFormat("EEEE, dd.MMMM yyyy").format(day);
	}
	
	public Date getDay() {
		return day;
	}

	public void setDay(Date day) {
		this.day = day;
	}

	public String getCalendarEvents() {
		return calendarEvents;
	}

	/**
	 * @return the users
	 */
	public List<UserBean> getUsers() {
		return users;
	}

	/**
	 * @return the downloadBean
	 */
	public DownloadBean getDownloadBean() {
		return downloadBean;
	}
}
